/*
 * eventdb.cpp - asynchronous I/O event database
 * Copyright (C) 2001, 2002  Justin Karneges
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 */

#include "eventdb.h"

#define FAKEDELAY 0

#include <QVector>
#include <QFileInfo>
#include <QDir>
#include <QTimer>
#include <QTextStream>
#include <QDateTime>
#include <QSqlError>
#include <QSqlDriver>
#include <qjson/parser.h>
#include "qjson/serializer.h"

#include "common.h"
#include "applicationinfo.h"
#include "psievent.h"
#include "psicontactlist.h"
#include "jidutil.h"
#include "historyimp.h"

using namespace XMPP;

//----------------------------------------------------------------------------
// EDBItem
//----------------------------------------------------------------------------
EDBItem::EDBItem(const PsiEvent::Ptr &event, const QString &id)
{
	e = event;
	v_id = id;
}

EDBItem::~EDBItem()
{
}

PsiEvent::Ptr EDBItem::event() const
{
	return e;
}

const QString & EDBItem::id() const
{
	return v_id;
}

//----------------------------------------------------------------------------
// EDBHandle
//----------------------------------------------------------------------------
class EDBHandle::Private
{
public:
	Private() {}

	EDB *edb;
	int beginRow_;
	EDBResult r;
	bool busy;
	bool writeSuccess;
	int listeningFor;
	int lastRequestType;
};

EDBHandle::EDBHandle(EDB *edb)
:QObject(0)
{
	d = new Private;
	d->edb = edb;
	d->beginRow_ = 0;
	d->busy = false;
	d->writeSuccess = false;
	d->listeningFor = -1;
	d->lastRequestType = Read;

	d->edb->reg(this);
}

EDBHandle::~EDBHandle()
{
	d->edb->unreg(this);

	delete d;
}

void EDBHandle::get(const QString &accId, const XMPP::Jid &jid, const QDateTime date, int direction, int begin, int len)
{
	d->busy = true;
	d->lastRequestType = Read;
	d->listeningFor = d->edb->op_get(accId, jid, date, direction, begin, len);
}

void EDBHandle::find(const QString &accId, const QString &str, const XMPP::Jid &jid, const QDateTime date, int direction)
{
	d->busy = true;
	d->lastRequestType = Read;
	d->listeningFor = d->edb->op_find(accId, str, jid, date, direction);
}

void EDBHandle::append(const QString &accId, const Jid &j, const PsiEvent::Ptr &e, int type)
{
	d->busy = true;
	d->lastRequestType = Write;
	d->listeningFor = d->edb->op_append(accId, j, e, type);
}

void EDBHandle::erase(const QString &accId, const Jid &j)
{
	d->busy = true;
	d->lastRequestType = Erase;
	d->listeningFor = d->edb->op_erase(accId, j);
}

bool EDBHandle::busy() const
{
	return d->busy;
}

const EDBResult EDBHandle::result() const
{
	return d->r;
}

bool EDBHandle::writeSuccess() const
{
	return d->writeSuccess;
}

void EDBHandle::edb_resultReady(EDBResult r)
{
	d->busy = false;
	d->r = r;
	d->listeningFor = -1;
	finished();
}

void EDBHandle::edb_writeFinished(bool b)
{
	d->busy = false;
	d->writeSuccess = b;
	d->listeningFor = -1;
	finished();
}

int EDBHandle::listeningFor() const
{
	return d->listeningFor;
}

int EDBHandle::lastRequestType() const
{
	return d->lastRequestType;
}

int EDBHandle::beginRow() const
{
	return d->beginRow_;
}


//----------------------------------------------------------------------------
// EDB
//----------------------------------------------------------------------------
class EDB::Private
{
public:
	Private() {}

	QList<EDBHandle*> list;
	int reqid_base;
	PsiCon *psi;
};

EDB::EDB(PsiCon *psi)
{
	d = new Private;
	d->reqid_base = 0;
	d->psi = psi;
}

EDB::~EDB()
{
	qDeleteAll(d->list);
	d->list.clear();
	delete d;
}

int EDB::genUniqueId() const
{
	return d->reqid_base++;
}

void EDB::reg(EDBHandle *h)
{
	d->list.append(h);
}

void EDB::unreg(EDBHandle *h)
{
	d->list.removeAll(h);
}

int EDB::op_get(const QString &accId, const Jid &jid, const QDateTime date, int direction, int start, int len)
{
	return get(accId, jid, date, direction, start, len);
}

int EDB::op_find(const QString &accId, const QString &str, const Jid &j, const QDateTime date, int direction)
{
	return find(accId, str, j, date, direction);
}

int EDB::op_append(const QString &accId, const Jid &j, const PsiEvent::Ptr &e, int type)
{
	return append(accId, j, e, type);
}

int EDB::op_erase(const QString &accId, const Jid &j)
{
	return erase(accId, j);
}

void EDB::resultReady(int req, EDBResult r, int begin_row)
{
	// deliver
	foreach(EDBHandle* h, d->list) {
		if(h->listeningFor() == req) {
			h->d->beginRow_ = begin_row;
			h->edb_resultReady(r);
			return;
		}
	}
}

void EDB::writeFinished(int req, bool b)
{
	// deliver
	foreach(EDBHandle* h, d->list) {
		if(h->listeningFor() == req) {
			h->edb_writeFinished(b);
			return;
		}
	}
}

inline PsiCon *EDB::psi()
{
	return d->psi;
}


//----------------------------------------------------------------------------
// EDBFlatFile
//----------------------------------------------------------------------------
struct item_file_req
{
	Jid j;
	int type; // 0 = latest, 1 = oldest, 2 = random, 3 = write
	int start;
	int len;
	int dir;
	int id;
	QDateTime date;
	QString findStr;
	PsiEvent::Ptr event;

	enum Type {
		Type_get,
		Type_append,
		Type_find,
		Type_erase
	};
};

class EDBFlatFile::Private
{
public:
	Private() {}

	QList<File*> flist;
	QList<item_file_req*> rlist;
};

EDBFlatFile::EDBFlatFile(PsiCon *psi)
:EDB(psi)
{
	d = new Private;
}

EDBFlatFile::~EDBFlatFile()
{
	qDeleteAll(d->rlist);
	qDeleteAll(d->flist);
	d->flist.clear();

	delete d;
}

int EDBFlatFile::get(const QString &/*accId*/, const Jid &j, const QDateTime date, int direction, int start, int len)
{
	item_file_req *r = new item_file_req;
	r->j = j;
	r->type = item_file_req::Type_get;
	r->start = start;
	r->len = len < 1 ? 1: len;
	r->dir = direction;
	r->date = date;
	r->id = genUniqueId();
	d->rlist.append(r);

	QTimer::singleShot(FAKEDELAY, this, SLOT(performRequests()));
	return r->id;
}

int EDBFlatFile::find(const QString &/*accId*/, const QString &str, const Jid &j, const QDateTime date, int direction)
{
	item_file_req *r = new item_file_req;
	r->j = j;
	r->type = item_file_req::Type_find;
	r->len = 1;
	r->dir = direction;
	r->findStr = str;
	r->date = date;
	r->id = genUniqueId();
	d->rlist.append(r);

	QTimer::singleShot(FAKEDELAY, this, SLOT(performRequests()));
	return r->id;
}

int EDBFlatFile::append(const QString &/*accId*/, const Jid &j, const PsiEvent::Ptr &e, int type)
{
	if (type != EDB::Contact)
		return 0;
	item_file_req *r = new item_file_req;
	r->j = j;
	r->type = item_file_req::Type_append;
	r->event = e;
	if ( !r->event ) {
		qWarning("EDBFlatFile::append(): Attempted to append incompatible type.");
		delete r;
		return 0;
	}
	r->id = genUniqueId();
	d->rlist.append(r);

	QTimer::singleShot(FAKEDELAY, this, SLOT(performRequests()));
	return r->id;
}

int EDBFlatFile::erase(const QString &/*accId*/, const Jid &j)
{
	item_file_req *r = new item_file_req;
	r->j = j;
	r->type = item_file_req::Type_erase;
	r->id = genUniqueId();
	d->rlist.append(r);

	QTimer::singleShot(FAKEDELAY, this, SLOT(performRequests()));
	return r->id;
}

QList<EDB::ContactItem> EDBFlatFile::contacts(const QString &/*accId*/, int type)
{
	return File::contacts(type);
}

quint64 EDBFlatFile::eventsCount(const QString &accId, const XMPP::Jid &jid)
{
	quint64 res = 0;
	if (!jid.isEmpty())
		res = ensureFile(jid)->total();
	else
		foreach (const ContactItem &ci, contacts(accId, Contact))
			res += ensureFile(ci.jid)->total();
	return res;
}

EDBFlatFile::File *EDBFlatFile::findFile(const Jid &j) const
{
	foreach(File* i, d->flist) {
		if(i->j.compare(j, false))
			return i;
	}
	return 0;
}

EDBFlatFile::File *EDBFlatFile::ensureFile(const Jid &j)
{
	File *i = findFile(j);
	if(!i) {
		i = new File(Jid(j.bare()));
		connect(i, SIGNAL(timeout()), SLOT(file_timeout()));
		d->flist.append(i);
	}
	return i;
}

bool EDBFlatFile::deleteFile(const Jid &j)
{
	File *i = findFile(j);

	QString fname;

	if (i) {
		fname = i->fname;
		d->flist.removeAll(i);
		delete i;
	}
	else {
		fname = File::jidToFileName(j);
	}

	QFileInfo fi(fname);
	if(fi.exists()) {
		QDir dir = fi.dir();
		return dir.remove(fi.fileName());
	}
	else
		return true;
}

void EDBFlatFile::performRequests()
{
	if(d->rlist.isEmpty())
		return;

	item_file_req *r = d->rlist.takeFirst();

	File *f = ensureFile(r->j);
	int type = r->type;
	if(type == item_file_req::Type_get) {

		int direction = r->dir;
		int id = f->getId(r->date, direction, r->start);

		int len;
		if(direction == Forward) {
			if(id + r->len > f->total())
				len = f->total() - id;
			else
				len = r->len;
		}
		else {
			if((id+1) - r->len < 0)
				len = id+1;
			else
				len = r->len;
		}

		EDBResult result;
		int startId = id;
		for(int n = 0; n < len; ++n) {
			PsiEvent::Ptr e(f->get(id));
			if(e) {
				EDBItemPtr ei = EDBItemPtr(new EDBItem(e, QString::number(id)));
				result.append(ei);
			}

			if(direction == Forward)
				++id;
			else
				--id;
		}
		if (direction == Backward)
			startId = id + 1;
		resultReady(r->id, result, startId);
	}
	else if(type == item_file_req::Type_append) {
		writeFinished(r->id, f->append(r->event));
	}
	else if(type == item_file_req::Type_find) {
		int id = f->getId(r->date, r->dir, r->start);
		EDBResult result;
		while(1) {
			PsiEvent::Ptr e(f->get(id));
			if(!e)
				break;

			if(e->type() == PsiEvent::Message) {
				MessageEvent::Ptr me = e.staticCast<MessageEvent>();
				const Message &m = me->message();
				if(m.body().indexOf(r->findStr, 0, Qt::CaseInsensitive) != -1) {
					EDBItemPtr ei = EDBItemPtr(new EDBItem(e, QString::number(id)));
					result.append(ei);
					//commented line below to return ALL(instead of just first) messages that contain findStr
					//break;
				}
			}

			if(r->dir == Forward)
				++id;
			else
				--id;
		}
		resultReady(r->id, result, 0);
	}

	else if(type == item_file_req::Type_erase) {
		writeFinished(r->id, deleteFile(f->j));
	}

	else {
		qWarning("EDBFlatFile::performRequests(): Invalid type.");
	}

	delete r;
}

void EDBFlatFile::file_timeout()
{
	File *i = (File *)sender();
	d->flist.removeAll(i);
	i->deleteLater();
}


//----------------------------------------------------------------------------
// EDBFlatFile::File
//----------------------------------------------------------------------------
class EDBFlatFile::File::Private
{
public:
	Private() {}

	QVector<quint64> index;
	bool indexed;
};

EDBFlatFile::File::File(const Jid &_j)
{
	d = new Private;
	d->indexed = false;

	j = _j;
	valid = false;
	t = new QTimer(this);
	connect(t, SIGNAL(timeout()), SLOT(timer_timeout()));

	//printf("[EDB opening -- %s]\n", j.full().latin1());
	fname = jidToFileName(_j);
	f.setFileName(fname);
	valid = f.open(QIODevice::ReadWrite);

	touch();
}

EDBFlatFile::File::~File()
{
	if(valid)
		f.close();
	//printf("[EDB closing -- %s]\n", j.full().latin1());

	delete d;
}

QString EDBFlatFile::File::jidToFileName(const XMPP::Jid &j)
{
	return ApplicationInfo::historyDir() + "/" + strToFileName(JIDUtil::encode(j.bare()).toLower());
}

QString EDBFlatFile::File::strToFileName(const QString &s)
{
	QFileInfo fi(s);
	return fi.fileName() + ".history";
}

QList<EDB::ContactItem> EDBFlatFile::File::contacts(int type)
{
	QList<ContactItem> res;
	if (type == EDB::Contact) {
		QDir dir(ApplicationInfo::historyDir() + "/");
		QFileInfoList flist = dir.entryInfoList(QStringList(strToFileName("*")), QDir::Files);
		foreach (const QFileInfo &fi, flist) {
			XMPP::Jid jid(JIDUtil::decode(fi.completeBaseName()));
			if (jid.isValid())
				res.append(ContactItem("", jid));
		}
	}
	return res;
}

void EDBFlatFile::File::ensureIndex()
{
	if ( valid && !d->indexed ) {
		if (f.isSequential()) {
			qWarning("EDBFlatFile::File::ensureIndex(): Can't index sequential files.");
			return;
		}

		f.reset(); // go to beginning
		d->index.clear();

		//printf(" file: %s\n", fname.latin1());
		// build index
		while(1) {
			quint64 at = f.pos();

			// locate a newline
			bool found = false;
			char c;
			while (f.getChar(&c)) {
				if (c == '\n') {
					found = true;
					break;
				}
			}

			if(!found)
				break;

			int oldsize = d->index.size();
			d->index.resize(oldsize+1);
			d->index[oldsize] = at;
		}

		d->indexed = true;
	}
	else {
		//printf(" file: can't open\n");
	}

	//printf(" messages: %d\n\n", d->index.size());
}

int EDBFlatFile::File::total() const
{
	((EDBFlatFile::File *)this)->ensureIndex();
	return d->index.size();
}

int EDBFlatFile::File::getId(QDateTime &date, int dir, int offset)
{
	if (date.isNull()) {
		if (dir == EDBFlatFile::Forward)
			return offset;
		if (offset >= total())
			return 0;
		return total() - offset - 1;
	}
	ensureIndex();
	int id = findIdByNearDate(date, 0, total()-1);
	if (id != -1) {
		QDateTime fDate = getLineDate(getLine(id));
		if (dir == EDBFlatFile::Forward) {
			if (fDate < date)
				++id;
			id += offset;
		} else {
			if (fDate > date)
				--id;
			id -= offset;
		}
		if (id >= total())
			id = total() - 1;
		else if (id < 0)
			id = 0;
	}
	return id;
}

int EDBFlatFile::File::findIdByNearDate(QDateTime &date, int start, int end)
{
	if (start == end)
		return start;

	int middle = (end - start) / 2;
	const QDateTime mDate = getLineDate(getLine(start + middle));
	if (!mDate.isValid())
		return -1;

	if (mDate == date)
		return start + middle;
	if (middle == 0) {
		const QDateTime mDate2 = getLineDate(getLine(end));
		if (!mDate2.isValid())
			return -1;
		return (abs(date.secsTo(mDate)) < abs(date.secsTo(mDate2))) ? start : end;
	}
	if (mDate > date)
		return findIdByNearDate(date, start, start + middle);
	return findIdByNearDate(date, start + middle, end);
}

void EDBFlatFile::File::touch()
{
	t->start(30000);
}

void EDBFlatFile::File::timer_timeout()
{
	timeout();
}

QString EDBFlatFile::File::getLine(int id)
{
	touch();

	if(!valid)
		return QString();

	ensureIndex();
	if(id < 0 || id >= (int)d->index.size())
		return QString();

	f.seek(d->index[id]);

	QTextStream t;
	t.setDevice(&f);
	t.setCodec("UTF-8");
	return t.readLine();
}

PsiEvent::Ptr EDBFlatFile::File::get(int id)
{
	QString line = getLine(id);
	if (line.isEmpty())
		return PsiEvent::Ptr();
	return lineToEvent(line);
}

bool EDBFlatFile::File::append(const PsiEvent::Ptr &e)
{
	touch();

	if(!valid)
		return false;

	QString line = eventToLine(e);
	if(line.isEmpty())
		return false;

	f.seek(f.size());
	quint64 at = f.pos();

	QTextStream t;
	t.setDevice(&f);
	t.setCodec("UTF-8");
	t << line << endl;
	f.flush();

	if ( d->indexed ) {
		int oldsize = d->index.size();
		d->index.resize(oldsize+1);
		d->index[oldsize] = at;
	}

	return true;
}

QDateTime EDBFlatFile::File::getLineDate(const QString &line) const
{
	int x1 = line.indexOf('|') + 1;
	int x2 = line.indexOf('|', x1);
	return QDateTime::fromString(line.mid(x1, x2 - x1), Qt::ISODate);
}

PsiEvent::Ptr EDBFlatFile::File::lineToEvent(const QString &line)
{
	// -- read the line --
	QString sTime, sType, sOrigin, sFlags, sText, sSubj, sUrl, sUrlDesc;
	int x1, x2;
	x1 = line.indexOf('|') + 1;

	x2 = line.indexOf('|', x1);
	sTime = line.mid(x1, x2-x1);
	x1 = x2 + 1;

	x2 = line.indexOf('|', x1);
	sType = line.mid(x1, x2-x1);
	x1 = x2 + 1;

	x2 = line.indexOf('|', x1);
	sOrigin = line.mid(x1, x2-x1);
	x1 = x2 + 1;

	x2 = line.indexOf('|', x1);
	sFlags = line.mid(x1, x2-x1);
	x1 = x2 + 1;

	// check for extra fields
	if(sFlags[1] != '-') {
		int subflags = QString(sFlags[1]).toInt(NULL,16);

		// have subject?
		if(subflags & 1) {
			x2 = line.indexOf('|', x1);
			sSubj = line.mid(x1, x2-x1);
			x1 = x2 + 1;
		}
		// have url?
		if(subflags & 2) {
			x2 = line.indexOf('|', x1);
			sUrl = line.mid(x1, x2-x1);
			x1 = x2 + 1;
			x2 = line.indexOf('|', x1);
			sUrlDesc = line.mid(x1, x2-x1);
			x1 = x2 + 1;
		}
	}

	// body text is last
	sText = line.mid(x1);

	// -- read end --

	int type = sType.toInt();
	if(type == 0 || type == 1 || type == 4 || type == 5) {
		Message m;
		m.setTimeStamp(QDateTime::fromString(sTime, Qt::ISODate));
		if(type == 1)
			m.setType("chat");
		else if(type == 4)
			m.setType("error");
		else if(type == 5)
			m.setType("headline");
		else
			m.setType("");

		bool originLocal = (sOrigin == "to") ? true: false;
		m.setFrom(j);
		if(sFlags[0] == 'N')
			m.setBody(logdecode(sText));
		else
			m.setBody(logdecode(QString::fromUtf8(sText.toLatin1())));
		m.setSubject(logdecode(sSubj));

		QString url = logdecode(sUrl);
		if(!url.isEmpty())
			m.urlAdd(Url(url, logdecode(sUrlDesc)));
		m.setSpooled(true);

		MessageEvent::Ptr me(new MessageEvent(m, 0));
		me->setOriginLocal(originLocal);

		return me.staticCast<PsiEvent>();
	}
	else if(type == 2 || type == 3 || type == 6 || type == 7 || type == 8) {
		QString subType = "subscribe";
		if(type == 2) {
			// stupid "system message" from Psi <= 0.8.6
			// try to figure out what kind it REALLY is based on the text
			if(sText == tr("<big>[System Message]</big><br>You are now authorized."))
				subType = "subscribed";
			else if(sText == tr("<big>[System Message]</big><br>Your authorization has been removed!"))
				subType = "unsubscribed";
		}
		else if(type == 3)
			subType = "subscribe";
		else if(type == 6)
			subType = "subscribed";
		else if(type == 7)
			subType = "unsubscribe";
		else if(type == 8)
			subType = "unsubscribed";

		AuthEvent::Ptr ae(new AuthEvent(j, subType, 0));
		ae->setTimeStamp(QDateTime::fromString(sTime, Qt::ISODate));
		return ae.staticCast<PsiEvent>();
	}

	return PsiEvent::Ptr();
}

QString EDBFlatFile::File::eventToLine(const PsiEvent::Ptr &e)
{
	int subflags = 0;
	QString sTime, sType, sOrigin, sFlags;

	if(e->type() == PsiEvent::Message) {
		MessageEvent::Ptr me = e.staticCast<MessageEvent>();
		const Message &m = me->message();
		const UrlList urls = m.urlList();

		if(!m.subject().isEmpty())
			subflags |= 1;
		if(!urls.isEmpty())
			subflags |= 2;

		sTime = m.timeStamp().toString(Qt::ISODate);
		int n = 0;
		if(m.type() == "chat")
			n = 1;
		else if(m.type() == "error")
			n = 4;
		else if(m.type() == "headline")
			n = 5;
		sType.setNum(n);
		sOrigin = e->originLocal() ? "to": "from";
		sFlags = "N---";

		if(subflags != 0)
			sFlags[1] = QString::number(subflags,16)[0];

		//  | date | type | To/from | flags | text
		QString line = "|" + sTime + "|" + sType + "|" + sOrigin + "|" + sFlags + "|";

		if(subflags & 1) {
			line += logencode(m.subject()) + "|";
		}
		if(subflags & 2) {
			const Url &url = urls.first();
			line += logencode(url.url()) + "|";
			line += logencode(url.desc()) + "|";
		}
		line += logencode(m.body());

		return line;
	}
	else if(e->type() == PsiEvent::Auth) {
		AuthEvent::Ptr ae = e.staticCast<AuthEvent>();
		sTime = ae->timeStamp().toString(Qt::ISODate);
		QString subType = ae->authType();
		int n = 0;
		if(subType == "subscribe")
			n = 3;
		else if(subType == "subscribed")
			n = 6;
		else if(subType == "unsubscribe")
			n = 7;
		else if(subType == "unsubscribed")
			n = 8;
		sType.setNum(n);
		sOrigin = e->originLocal() ? "to": "from";
		sFlags = "N---";

		//  | date | type | To/from | flags | text
		QString line = "|" + sTime + "|" + sType + "|" + sOrigin + "|" + sFlags + "|";
		line += logencode(subType);

		return line;
	}

	return "";
}

//----------------------------------------------------------------------------
// EDBSqLite
//----------------------------------------------------------------------------

EDBSqLite::EDBSqLite(PsiCon *psi) : EDB(psi),
	transactionsCounter(0),
	lastCommitTime(QDateTime::currentDateTime()),
	commitTimer(NULL),
	mirror_(NULL)
{
	status = NotActive;
	QString path = ApplicationInfo::historyDir() + "/history.db";
	QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "history");
	db.setDatabaseName(path);
	if (!db.open()) {
		qWarning("EDBSqLite::EDBSqLite(): Can't open base.\n" + db.lastError().text().toLatin1());
		return;
	}
	QSqlQuery query(db);
	query.exec("PRAGMA foreign_keys = ON;");
	setInsertingMode(Normal);
	if (db.tables(QSql::Tables).size() == 0) {
		// no tables found.
		if (db.transaction()) {
			query.exec("CREATE TABLE `system` ("
				"`key` TEXT, "
				"`value` TEXT"
				");");
			query.exec("CREATE TABLE `accounts` ("
				"`id` TEXT, "
				"`lifetime` INTEGER"
				");");
			query.exec("CREATE TABLE `contacts` ("
				"`id` INTEGER NOT NULL PRIMARY KEY ASC, "
				"`acc_id` TEXT, "
				"`type` INTEGER, "
				"`jid` TEXT, "
				"`lifetime` INTEGER"
				");");
			query.exec("CREATE TABLE `events` ("
				"`id` INTEGER NOT NULL PRIMARY KEY ASC, "
				"`contact_id` INTEGER NOT NULL REFERENCES `contacts`(`id`) ON DELETE CASCADE, "
				"`resource` TEXT, "
				"`date` TEXT, "
				"`type` INTEGER, "
				"`direction` INTEGER, "
				"`subject` TEXT, "
				"`m_text` TEXT, "
				"`lang` TEXT, "
				"`extra_data` TEXT"
				");");
			query.exec("CREATE INDEX `key` ON `system` (`key`);");
			query.exec("CREATE INDEX `jid` ON `contacts` (`jid`);");
			query.exec("CREATE INDEX `contact_id` ON `events` (`contact_id`);");
			query.exec("CREATE INDEX `date` ON `events` (`date`);");
			if (db.commit()) {
				status = Commited;
				setStorageParam("version", "0.1");
				setStorageParam("import_start", "yes");
			}
		}
	}
	else
		status = Commited;
}

EDBSqLite::~EDBSqLite()
{
	commit();
	{
		QSqlDatabase db = QSqlDatabase::database("history", false);
		if (db.isOpen())
			db.close();
	}
	QSqlDatabase::removeDatabase("history");
}

bool EDBSqLite::init()
{
	if (status == NotActive)
		return false;

	if (!getStorageParam("import_start").isEmpty()) {
		if (!importExecute()) {
			status = NotActive;
			return false;
		}
	}

	setMirror(new EDBFlatFile(psi()));
	return true;
}

int EDBSqLite::get(const QString &accId, const XMPP::Jid &jid, QDateTime date, int direction, int start, int len)
{
	item_query_req *r = new item_query_req;
	r->accId = accId;
	r->j     = jid;
	r->type  = item_query_req::Type_get;
	r->start = start;
	r->len   = len < 1 ? 1 : len;
	r->dir   = direction;
	r->date  = date;
	r->id    = genUniqueId();
	rlist.append(r);

	QTimer::singleShot(FAKEDELAY, this, SLOT(performRequests()));
	return r->id;
}

int EDBSqLite::find(const QString &accId, const QString &str, const XMPP::Jid &jid, const QDateTime date, int direction)
{
	item_query_req *r = new item_query_req;
	r->accId   = accId;
	r->j       = jid;
	r->type    = item_query_req::Type_find;
	r->len     = 1;
	r->dir     = direction;
	r->findStr = str;
	r->date    = date;
	r->id      = genUniqueId();
	rlist.append(r);

	QTimer::singleShot(FAKEDELAY, this, SLOT(performRequests()));
	return r->id;
}

int EDBSqLite::append(const QString &accId, const XMPP::Jid &jid, const PsiEvent::Ptr &e, int type)
{
	item_query_req *r = new item_query_req;
	r->accId   = accId;
	r->j       = jid;
	r->jidType = type;
	r->type    = item_file_req::Type_append;
	r->event   = e;
	if ( !r->event ) {
		qWarning("EDBSqLite::append(): Attempted to append incompatible type.");
		delete r;
		return 0;
	}
	r->id      = genUniqueId();
	rlist.append(r);

	QTimer::singleShot(FAKEDELAY, this, SLOT(performRequests()));

	if (mirror_)
		mirror_->append(accId, jid, e, type);

	return r->id;
}

int EDBSqLite::erase(const QString &accId, const XMPP::Jid &jid)
{
	item_query_req *r = new item_query_req;
	r->accId = accId;
	r->j     = jid;
	r->type  = item_query_req::Type_erase;
	r->id    = genUniqueId();
	rlist.append(r);

	QTimer::singleShot(FAKEDELAY, this, SLOT(performRequests()));
	return r->id;
}

QList<EDB::ContactItem> EDBSqLite::contacts(const QString &accId, int type)
{
	QList<ContactItem> res;
	EDBSqLite::PreparedQuery *query = queryes.getPreparedQuery(QueryContactsList, accId.isEmpty(), true);
	query->bindValue(":type", type);
	if (!accId.isEmpty())
		query->bindValue(":acc_id", accId);
	if (query->exec()) {
		while (query->next()) {
			const QSqlRecord &rec = query->record();
			res.append(ContactItem(rec.value("acc_id").toString(), XMPP::Jid(rec.value("jid").toString())));
		}
		query->freeResult();
	}
	return res;
}

quint64 EDBSqLite::eventsCount(const QString &accId, const XMPP::Jid &jid)
{
	quint64 res = 0;
	bool fAccAll  = accId.isEmpty();
	bool fContAll = jid.isEmpty();
	EDBSqLite::PreparedQuery *query = queryes.getPreparedQuery(QueryRowCount, fAccAll, fContAll);
	if (!fAccAll)
		query->bindValue(":acc_id", accId);
	if (!fContAll)
		query->bindValue(":jid", jid.full());
	if (query->exec()) {
		if (query->next())
			res = query->record().value("count").toULongLong();
		query->freeResult();
	}
	return res;
}

QString EDBSqLite::getStorageParam(const QString &key)
{
	QSqlQuery query(QSqlDatabase::database("history"));
	query.prepare("SELECT `value` FROM `system` WHERE `key` = :key;");
	query.bindValue(":key", key);
	if (query.exec() && query.next())
		return query.record().value("value").toString();
	return QString();
}

void EDBSqLite::setStorageParam(const QString &key, const QString &val)
{
	transaction(true);
	QSqlQuery query(QSqlDatabase::database("history"));
	if (val.isEmpty()) {
		query.prepare("DELETE FROM `system` WHERE `key` = :key;");
		query.bindValue(":key", key);
		query.exec();
	}
	else {
		query.prepare("SELECT COUNT(*) AS `count` FROM `system` WHERE `key` = :key;");
		query.bindValue(":key", key);
		if (query.exec() && query.next() && query.record().value("count").toULongLong() != 0) {
			query.prepare("UPDATE `system` SET `value` = :val WHERE `key` = :key;");
			query.bindValue(":key", key);
			query.bindValue(":val", val);
			query.exec();
		}
		else {
			query.prepare("INSERT INTO `system` (`key`, `value`) VALUES (:key, :val);");
			query.bindValue(":key", key);
			query.bindValue(":val", val);
			query.exec();
		}
	}
	commit();
}

void EDBSqLite::setInsertingMode(InsertMode mode)
{
	// in the case of a flow of new records
	if (mode == Import) {
		// Commit after 10000 inserts and every 5 seconds
		maxUncommitedRecs = 10000;
		maxUncommitedSecs = 5;
	} else {
		// Commit after 3 inserts and every 1 second
		maxUncommitedRecs = 3;
		maxUncommitedSecs = 1;
	}
	// Commit if there were no new additions for 1 second
	commitByTimeoutSecs = 1;
	//--
	commit();
}

void EDBSqLite::setMirror(EDBFlatFile *mirr)
{
	if (mirr != mirror_) {
		if (mirror_)
			delete mirror_;
		mirror_ = mirr;
	}
}

EDBFlatFile *EDBSqLite::mirror() const
{
	return mirror_;
}

void EDBSqLite::performRequests()
{
	if (rlist.isEmpty())
		return;

	item_query_req *r = rlist.takeFirst();
	const int type = r->type;

	if (type == item_query_req::Type_append) {
		bool b = appendEvent(r->accId, r->j, r->event, r->jidType);
		writeFinished(r->id, b);
	}

	else if (type == item_query_req::Type_get) {
		commit();
		bool fContAll = r->j.isEmpty();
		bool fAccAll  = r->accId.isEmpty();
		QueryType queryType;
		if (r->date.isNull()) {
			if (r->dir == Forward)
				queryType = QueryOldest;
			else
				queryType = QueryLatest;
		} else {
			if (r->dir == Backward)
				queryType = QueryDateBackward;
			else
				queryType = QueryDateForward;
		}
		EDBSqLite::PreparedQuery *query = queryes.getPreparedQuery(queryType, fAccAll, fContAll);
		if (!fContAll)
			query->bindValue(":jid", r->j.full());
		if (!fAccAll)
			query->bindValue(":acc_id", r->accId);
		if (!r->date.isNull())
			query->bindValue(":date", r->date);
		query->bindValue(":start", r->start);
		query->bindValue(":cnt", r->len);
		EDBResult result;
		if (query->exec()) {
			while (query->next()) {
				PsiEvent::Ptr e(getEvent(query->record()));
				if (e) {
					QString id = query->record().value("id").toString();
					result.append(EDBItemPtr(new EDBItem(e, id)));
				}
			}
			query->freeResult();
		}
		int beginRow;
		if (r->dir == Forward && r->date.isNull()) {
			beginRow = r->start;
		} else {
			int cnt = rowCount(r->accId, r->j, r->date);
			if (r->dir == Backward) {
				beginRow = cnt - r->len + 1;
				if (beginRow < 0)
					beginRow = 0;
			} else {
				beginRow = cnt + 1;
			}
		}
		resultReady(r->id, result, beginRow);

	} else if(type == item_file_req::Type_find) {
		commit();
		bool fContAll = r->j.isEmpty();
		bool fAccAll  = r->accId.isEmpty();
		EDBSqLite::PreparedQuery *query = queryes.getPreparedQuery(QueryFindText, fAccAll, fContAll);
		if (!fContAll)
			query->bindValue(":jid", r->j.bare());
		if (!fAccAll)
			query->bindValue(":acc_id", r->accId);
		EDBResult result;
		if (query->exec()) {
			QString str = r->findStr.toLower();
			while (query->next()) {
				const QSqlRecord rec = query->record();
				if (!rec.value("m_text").toString().toLower().contains(str, Qt::CaseSensitive))
					continue;
				PsiEvent::Ptr e(getEvent(rec));
				if (e) {
					QString id = rec.value("id").toString();
					EDBItemPtr eip = EDBItemPtr(new EDBItem(e, id));
					result.append(eip);
				}
			}
			query->freeResult();
		}
		resultReady(r->id, result, 0);

	} else if(type == item_file_req::Type_erase) {
		writeFinished(r->id, eraseHistory(r->accId, r->j));
	}

	delete r;
}

bool EDBSqLite::appendEvent(const QString &accId, const XMPP::Jid &jid, const PsiEvent::Ptr &e, int jidType)
{
	QSqlDatabase db = QSqlDatabase::database("history");
	const qint64 contactId = ensureJidRowId(accId, jid, jidType);
	if (contactId == 0)
		return false;

	QDateTime dTime;
	int nType = 0;

	if (e->type() == PsiEvent::Message) {
		MessageEvent::Ptr me = e.staticCast<MessageEvent>();
		const Message &m = me->message();
		dTime = m.timeStamp();
		if (m.type() == "chat")
			nType = 1;
		else if(m.type() == "error")
			nType = 4;
		else if(m.type() == "headline")
			nType = 5;

	} else if (e->type() == PsiEvent::Auth) {
		AuthEvent::Ptr ae = e.staticCast<AuthEvent>();
		dTime = ae->timeStamp();
		QString subType = ae->authType();
		if(subType == "subscribe")
			nType = 3;
		else if(subType == "subscribed")
			nType = 6;
		else if(subType == "unsubscribe")
			nType = 7;
		else if(subType == "unsubscribed")
			nType = 8;
	} else
		return false;

	int nDirection = e->originLocal() ? 1 : 2;
	if (!transaction(false))
		return false;

	PreparedQuery *query = queryes.getPreparedQuery(QueryInsertEvent, false, false);
	query->bindValue(":contact_id", contactId);
	query->bindValue(":resource", (jidType != GroupChatContact) ? jid.resource() : "");
	query->bindValue(":date", dTime);
	query->bindValue(":type", nType);
	query->bindValue(":direction", nDirection);
	if (nType == 0 || nType == 1 || nType == 4 || nType == 5) {
		MessageEvent::Ptr me = e.staticCast<MessageEvent>();
		const Message &m = me->message();
		QString lang = m.lang();
		query->bindValue(":subject", m.subject(lang));
		query->bindValue(":m_text", m.body(lang));
		query->bindValue(":lang", lang);
		QString extraData;
		const UrlList &urls = m.urlList();
		if (!urls.isEmpty()) {
			QVariantMap xepList;
			QVariantList urlList;
			foreach (const Url &url, urls)
				if (!url.url().isEmpty()) {
					QVariantList urlItem;
					urlItem.append(QVariant(url.url()));
					if (!url.desc().isEmpty())
						urlItem.append(QVariant(url.desc()));
					urlList.append(QVariant(urlItem));
				}
			xepList["jabber:x:oob"] = QVariant(urlList);
			QJson::Serializer serializer;
			extraData = QString::fromUtf8(serializer.serialize(xepList));
		}
		query->bindValue(":extra_data", extraData);
	}
	else {
		query->bindValue(":subject", QVariant(QVariant::String));
		query->bindValue(":m_text", QVariant(QVariant::String));
		query->bindValue(":lang", QVariant(QVariant::String));
		query->bindValue(":extra_data", QVariant(QVariant::String));
	}
	bool res = query->exec();
	return res;
}

PsiEvent::Ptr EDBSqLite::getEvent(const QSqlRecord &record)
{
	PsiAccount *pa = psi()->contactList()->getAccount(record.value("acc_id").toString());

	int type = record.value("type").toInt();

	if(type == 0 || type == 1 || type == 4 || type == 5) {
		Message m;
		m.setTimeStamp(record.value("date").toDateTime());
		if(type == 1)
			m.setType("chat");
		else if(type == 4)
			m.setType("error");
		else if(type == 5)
			m.setType("headline");
		else
			m.setType("");
		m.setFrom(Jid(record.value("jid").toString()));
		QVariant text = record.value("m_text");
		if (!text.isNull()) {
			m.setBody(text.toString());
			m.setLang(record.value("lang").toString());
			m.setSubject(record.value("subject").toString());
		}
		m.setSpooled(true);
		QString extraStr = record.value("extra_data").toString();
		if (!extraStr.isEmpty()) {
			QJson::Parser parser;
			bool fOk;
			QVariantMap extraData = parser.parse(extraStr.toUtf8(), &fOk).toMap();
			if (fOk) {
				foreach (const QVariant &urlItem, extraData["jabber:x:oob"].toList()) {
					QVariantList itemList = urlItem.toList();
					if (!itemList.isEmpty()) {
						QString url = itemList.at(0).toString();
						QString desc;
						if (itemList.size() > 1)
							desc = itemList.at(1).toString();
						m.urlAdd(Url(url, desc));
					}
				}
			}
		}
		MessageEvent::Ptr me(new MessageEvent(m, pa));
		me->setOriginLocal((record.value("direction").toInt() == 1));
		return me.staticCast<PsiEvent>();
	}

	if(type == 2 || type == 3 || type == 6 || type == 7 || type == 8) {
		QString subType = "subscribe";
		// if(type == 2) { // Not used (stupid "system message" from Psi <= 0.8.6)
		if(type == 3)
			subType = "subscribe";
		else if(type == 6)
			subType = "subscribed";
		else if(type == 7)
			subType = "unsubscribe";
		else if(type == 8)
			subType = "unsubscribed";

		AuthEvent::Ptr ae(new AuthEvent(Jid(record.value("jid").toString()), subType, pa));
		ae->setTimeStamp(record.value("date").toDateTime());
		return ae.staticCast<PsiEvent>();
	}
	return PsiEvent::Ptr();
}

qint64 EDBSqLite::ensureJidRowId(const QString &accId, const XMPP::Jid &jid, int type)
{
	if (jid.isEmpty())
		return 0;
	QString sJid = (type == GroupChatContact) ? jid.full() : jid.bare();
	QString sKey = accId + "|" + sJid;
	qint64 id = jidsCache.value(sKey, 0);
	if (id != 0)
		return id;

	EDBSqLite::PreparedQuery *query = queryes.getPreparedQuery(QueryJidRowId, false, false);
	query->bindValue(":jid", sJid);
	query->bindValue(":acc_id", accId);
	if (query->exec()) {
		if (query->first()) {
			id = query->record().value("id").toLongLong();
		} else {
			//
			QSqlQuery queryIns(QSqlDatabase::database("history"));
			queryIns.prepare("INSERT INTO `contacts` (`acc_id`, `type`, `jid`, `lifetime`)"
				" VALUES (:acc_id, :type, :jid, -1);");
			queryIns.bindValue(":acc_id", accId);
			queryIns.bindValue(":type", type);
			queryIns.bindValue(":jid", sJid);
			if (queryIns.exec()) {
				id  = queryIns.lastInsertId().toLongLong();
			}
		}
		query->freeResult();
		if (id != 0)
			jidsCache[sKey] = id;
	}
	return id;
}

int EDBSqLite::rowCount(const QString &accId, const XMPP::Jid &jid, QDateTime before)
{
	bool fAccAll  = accId.isEmpty();
	bool fContAll = jid.isEmpty();
	QueryType type;
	if (before.isNull())
		type = QueryRowCount;
	else
		type = QueryRowCountBefore;
	PreparedQuery *query = queryes.getPreparedQuery(type, fAccAll, fContAll);
	if (!fContAll)
		query->bindValue(":jid", jid.full());
	if (!fAccAll)
		query->bindValue(":acc_id", accId);
	if (!before.isNull())
		query->bindValue(":date", before);
	int res = 0;
	if (query->exec()) {
		if (query->next()) {
			res = query->record().value("count").toInt();
		}
		query->freeResult();
	}
	return res;
}

bool EDBSqLite::eraseHistory(const QString &accId, const XMPP::Jid &jid)
{
	bool res = false;
	if (!transaction(true))
		return false;

	if (accId.isEmpty() && jid.isEmpty()) {
		QSqlQuery query(QSqlDatabase::database("history"));
		//if (query.exec("DELETE FROM `events`;"))
			if (query.exec("DELETE FROM `contacts`;")) {
				jidsCache.clear();
				res = true;
			}
	}
	else {
		PreparedQuery *query = queryes.getPreparedQuery(QueryJidRowId, false, false);
		query->bindValue(":jid", jid.full());
		query->bindValue(":acc_id", accId);
		if (query->exec() && query->next()) {
			const qint64 id = query->record().value("id").toLongLong();
			QSqlQuery query2(QSqlDatabase::database("history"));
			query2.prepare("DELETE FROM `events` WHERE `contact_id` = :id;");
			query2.bindValue(":id", id);
			if (query2.exec()) {
				res = true;
				query2.prepare("DELETE FROM `contacts` WHERE `id` = :id AND `lifetime` = -1;");
				query2.bindValue(":id", id);
				if (query2.exec()) {
					if (query2.numRowsAffected() > 0)
							jidsCache.clear();
				} else
					res = false;
			}
		}
	}
	if (res)
		res = commit();
	else
		rollback();
	return res;
}

bool EDBSqLite::transaction(bool now)
{
	if (status == NotActive)
		return false;
	if (now || transactionsCounter >= maxUncommitedRecs
			|| lastCommitTime.secsTo(QDateTime::currentDateTime()) >= maxUncommitedSecs)
		if (!commit())
			return false;

	if (status == Commited) {
		if (!QSqlDatabase::database("history").transaction())
			return false;
		status = NotCommited;
	}
	++transactionsCounter;

	startAutocommitTimer();

	return true;
}

bool EDBSqLite::commit()
{
	if (status != NotActive) {
		if (status == Commited || QSqlDatabase::database("history").commit()) {
			transactionsCounter = 0;
			lastCommitTime = QDateTime::currentDateTime();
			status = Commited;
			stopAutocommitTimer();
			return true;
		}
	}
	return false;
}

bool EDBSqLite::rollback()
{
	if (status == NotCommited && QSqlDatabase::database("history").rollback()) {
		transactionsCounter = 0;
		lastCommitTime = QDateTime::currentDateTime();
		status = Commited;
		stopAutocommitTimer();
		return true;
	}
	return false;
}

void EDBSqLite::startAutocommitTimer()
{
	if (!commitTimer) {
		commitTimer = new QTimer(this);
		connect(commitTimer, SIGNAL(timeout()), this, SLOT(commit()));
		commitTimer->setSingleShot(true);
		commitTimer->setInterval(commitByTimeoutSecs * 1000);
	}
	commitTimer->start();
}

void EDBSqLite::stopAutocommitTimer()
{
	if (commitTimer && commitTimer->isActive())
		commitTimer->stop();
}

bool EDBSqLite::importExecute()
{
	bool res = true;
	HistoryImport *imp = new HistoryImport(psi());
	if (imp->isNeeded()) {
		if (imp->exec() != HistoryImport::ResultNormal) {
			res = false;
		}
	}
	delete imp;
	return res;
}

// ****************** class PreparedQueryes ********************

EDBSqLite::QueryStorage::QueryStorage()
{
}

EDBSqLite::QueryStorage::~QueryStorage()
{
	foreach (EDBSqLite::PreparedQuery *q, queryList.values()) {
		if (q)
			delete q;
	}
}

EDBSqLite::PreparedQuery *EDBSqLite::QueryStorage::getPreparedQuery(QueryType type, bool allAccounts, bool allContacts)
{
	QueryProperty queryProp(type, allAccounts, allContacts);
	EDBSqLite::PreparedQuery *q = queryList.value(queryProp, NULL);
	if (q != NULL)
		return q;

	q = new EDBSqLite::PreparedQuery(QSqlDatabase::database("history"));
	q->setForwardOnly(true);
	q->prepare(getQueryString(type, allAccounts, allContacts));
	queryList[queryProp] = q;
	return q;
}

EDBSqLite::PreparedQuery::PreparedQuery(QSqlDatabase db) : QSqlQuery(db)
{
}

QString EDBSqLite::QueryStorage::getQueryString(QueryType type, bool allAccounts, bool allContacts)
{
	QString queryStr;
	switch (type)
	{
		case QueryContactsList:
			queryStr = "SELECT `acc_id`, `jid` FROM `contacts` WHERE `type` = :type";
			if (!allAccounts)
				queryStr.append(" AND `acc_id` = :acc_id");
			queryStr.append(" ORDER BY `jid`;");
			break;
		case QueryLatest:
		case QueryOldest:
		case QueryDateBackward:
		case QueryDateForward:
			queryStr = "SELECT `acc_id`, `events`.`id`, `jid`, `date`, `events`.`type`, `direction`, `subject`, `m_text`, `lang`, `extra_data`"
				" FROM `events`, `contacts`"
				" WHERE `contacts`.`id` = `contact_id`";
			if (!allContacts)
				queryStr.append(" AND `jid` = :jid");
			if (!allAccounts)
				queryStr.append(" AND `acc_id` = :acc_id");
			if (type == QueryDateBackward)
				queryStr.append(" AND `date` < :date");
			else if (type == QueryDateForward)
				queryStr.append(" AND `date` >= :date");
			if (type == QueryLatest || type == QueryDateBackward)
				queryStr.append(" ORDER BY `date` DESC");
			else
				queryStr.append(" ORDER BY `date` ASC");
			queryStr.append(" LIMIT :start, :cnt;");
			break;
		case QueryRowCount:
		case QueryRowCountBefore:
			queryStr = "SELECT count(*) AS `count`"
				" FROM `events`, `contacts`"
				" WHERE `contacts`.`id` = `contact_id`";
			if (!allContacts)
				queryStr.append(" AND `jid` = :jid");
			if (!allAccounts)
				queryStr.append(" AND `acc_id` = :acc_id");
			if (type == QueryRowCountBefore)
				queryStr.append(" AND `date` < :date");
			queryStr.append(";");
			break;
		case QueryJidRowId:
			queryStr = "SELECT `id` FROM `contacts` WHERE `jid` = :jid AND acc_id = :acc_id;";
			break;
		case QueryFindText:
			queryStr = "SELECT `acc_id`, `events`.`id`, `jid`, `date`, `events`.`type`, `direction`, `subject`, `m_text`, `lang`, `extra_data`"
				" FROM `events`, `contacts`"
				" WHERE `contacts`.`id` = `contact_id`";
			if (!allContacts)
				queryStr.append(" AND `jid` = :jid");
			if (!allAccounts)
				queryStr.append(" AND `acc_id` = :acc_id");
			queryStr.append(" AND `m_text` IS NOT NULL");
			queryStr.append(" ORDER BY `date`;");
			break;
		case QueryInsertEvent:
			queryStr = "INSERT INTO `events` ("
				"`contact_id`, `resource`, `date`, `type`, `direction`, `subject`, `m_text`, `lang`, `extra_data`"
				") VALUES ("
				":contact_id, :resource, :date, :type, :direction, :subject, :m_text, :lang, :extra_data"
				");";
			break;
	}
	return queryStr;
}

uint qHash(const QueryProperty &struc)
{
	uint res = struc.type;
	res <<= 8;
	res |=  struc.allAccounts;
	res <<= 8;
	res |=  struc.allContacts;
	return res;
}
