/*
 * historydlg.cpp
 * Copyright (C) 2001-2010  Justin Karneges, Michail Pishchagin,
 * Piotr Okonski, Evgeny Khryukin
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 */

#include <QMessageBox>
#include <QScrollBar>
#include <QMenu>
#include <QProgressDialog>

#include "historydlg.h"
#include "psiaccount.h"
#include "psicon.h"
#include "psicontact.h"
#include "psiiconset.h"
#include "psioptions.h"
#include "coloropt.h"
#include "textutil.h"
#include "jidutil.h"
#include "fileutil.h"
#include "userlist.h"
#include "common.h"

static const QString geometryOption = "options.ui.history.size";

static QString getNext(QString *str)
{
	int n = 0;
	// skip leading spaces (but *do* return them later!)
	while(n < (int)str->length() && str->at(n).isSpace()) {
		++n;
	}
	if(n == (int)str->length()) {
		return QString::null;
	}
	// find end or next space
	while(n < (int)str->length() && !str->at(n).isSpace()) {
		++n;
	}
	QString result = str->mid(0, n);
	*str = str->mid(n);
	return result;
}

// wraps a string against a fixed width
static QStringList wrapString(const QString &str, int wid)
{
	QStringList lines;
	QString cur;
	QString tmp = str;
	//printf("parsing: [%s]\n", tmp.latin1());
	while(1) {
		QString word = getNext(&tmp);
		if(word == QString::null) {
			lines += cur;
			break;
		}
		//printf("word:[%s]\n", word.latin1());
		if(!cur.isEmpty()) {
			if((int)cur.length() + (int)word.length() > wid) {
				lines += cur;
				cur = "";
			}
		}
		if(cur.isEmpty()) {
			// trim the whitespace in front
			for(int n = 0; n < (int)word.length(); ++n) {
				if(!word.at(n).isSpace()) {
					if(n > 0) {
						word = word.mid(n);
					}
					break;
				}
			}
		}
		cur += word;
	}
	return lines;
}


class HistoryDlg::Private
{
public:
	Jid jid;
	PsiAccount *pa;
	PsiCon *psi;
	int begin_row, end_row;
	bool can_backward, can_forward;
	HistoryDlg::RequestType reqType;
	QString findStr;
	QDate date;
#ifndef HAVE_X11
	bool autoCopyText;
#endif
	bool formatting;
	bool emoticons;
	QString sentColor;
	QString receivedColor;
};

HistoryDlg::HistoryDlg(const Jid &jid, PsiAccount *pa)
	: AdvancedWidget<QDialog>(0, Qt::Window)
{
	ui_.setupUi(this);
	setAttribute(Qt::WA_DeleteOnClose);
	setModal(false);
	d = new Private;
	d->reqType = TypeNone;
	d->begin_row = 0;
	d->end_row = -1;
	d->psi = pa->psi();
	d->jid = jid;
	d->pa = pa;
	pa->dialogRegister(this, d->jid);

	//workaround calendar size
	int minWidth = ui_.calendar->minimumSizeHint().width();
	if(minWidth > ui_.calendar->maximumWidth()) {
		ui_.calendar->setMaximumWidth(minWidth);
		ui_.jidList->setMaximumWidth(minWidth);
		ui_.accountsBox->setMaximumWidth(minWidth);
		ui_.buttonRefresh->setMaximumWidth(minWidth);
	}

#ifndef Q_OS_MAC
	setWindowIcon(IconsetFactory::icon("psi/history").icon());
#endif
	ui_.tb_find->setIcon(IconsetFactory::icon("psi/search").icon());

	ui_.msgLog->setFont(fontForOption("options.ui.look.font.chat"));
	QFont f = fontForOption("options.ui.look.font.contactlist");
	ui_.jidList->setFont(f);
	ui_.jidList2->setFont(f);
	ui_.privList->setFont(f);
	ui_.advList->setFont(f);

	ui_.calendar->setFirstDayOfWeek(firstDayOfWeekFromLocale());

	connect(ui_.searchField, SIGNAL(returnPressed()), SLOT(findMessages()));
	connect(ui_.searchField, SIGNAL(textChanged(const QString)), SLOT(highlightBlocks(const QString)));
	connect(ui_.buttonPrevious, SIGNAL(released()), SLOT(getPrevious()));
	connect(ui_.buttonNext, SIGNAL(released()), SLOT(getNext()));
	connect(ui_.buttonRefresh, SIGNAL(released()), SLOT(refresh()));
	connect(ui_.jidList, SIGNAL(clicked(QModelIndex)), SLOT(openSelectedContact()));
	connect(ui_.jidList2, SIGNAL(clicked(QModelIndex)), SLOT(openSelectedContact()));
	connect(ui_.privList, SIGNAL(clicked(QModelIndex)), SLOT(openSelectedContact()));
	connect(ui_.advList, SIGNAL(clicked(QModelIndex)), SLOT(openSelectedContact()));
	connect(ui_.tb_find, SIGNAL(clicked()), SLOT(findMessages()));
	connect(ui_.buttonLastest, SIGNAL(released()), SLOT(getLatest()));
	connect(ui_.buttonEarliest, SIGNAL(released()), SLOT(getEarliest()));
	connect(ui_.calendar, SIGNAL(selectionChanged()), SLOT(getDate()));
	connect(ui_.calendar, SIGNAL(activated(QDate)), SLOT(getDate()));
#ifndef HAVE_X11	// linux has this feature built-in
	optionUpdated("options.ui.automatically-copy-selected-text");
	connect(ui_.msgLog, SIGNAL(selectionChanged()), SLOT(autoCopy()));
	connect(ui_.msgLog, SIGNAL(cursorPositionChanged()), SLOT(autoCopy()));
#endif
	optionUpdated("options.ui.look.colors.messages.sent");
	optionUpdated("options.ui.look.colors.messages.received");
	optionUpdated("options.ui.emoticons.use-emoticons");
	optionUpdated("options.ui.chat.legacy-formatting");
	connect(PsiOptions::instance(), SIGNAL(optionChanged(QString)), SLOT(optionUpdated(QString)));

	connect(pa, SIGNAL(removedContact(PsiContact*)), SLOT(removedContact(PsiContact*)));

	ui_.jidList->installEventFilter(this);
	ui_.jidList2->installEventFilter(this);
	ui_.privList->installEventFilter(this);
	ui_.advList->installEventFilter(this);

	listAccounts();
	loadContacts();

	setGeometryOptionPath(geometryOption);

	openSelectedContact();
}

HistoryDlg::~HistoryDlg()
{
	delete d;
}

bool HistoryDlg::eventFilter(QObject *obj, QEvent *e)
{
	if((obj == ui_.jidList || obj == ui_.jidList2 || obj == ui_.privList || obj == ui_.advList) && e->type() == QEvent::ContextMenu) {
		e->accept();
		QTimer::singleShot(0, this, SLOT(doMenu()));
		return true;
	}

	return QDialog::eventFilter(obj, e);
}

QFont HistoryDlg::fontForOption(const QString &option)
{
	QFont font;
	font.fromString(PsiOptions::instance()->getOption(option).toString());
	return font;
}

void HistoryDlg::changeAccount(const QString /*accountName*/)
{
	ui_.msgLog->clear();
	setButtons(false);
	d->jid = QString();
	loadContacts();
	setCurrentUserListItem();
	openSelectedContact();
}

void HistoryDlg::setCurrentUserListItem()
{
	switch (ui_.contactToolBox->currentIndex()) {
	case 0:
		ui_.jidList->setCurrentRow(0);
		break;
	case 1:
		ui_.jidList2->setCurrentRow(0);
		break;
	case 2:
		ui_.privList->setCurrentRow(0);
		break;
	case 3:
		ui_.advList->setCurrentRow(0);
		break;
	}
}

void HistoryDlg::listAccounts()
{
	ui_.accountsBox->addItem(IconsetFactory::icon("psi/account").icon(), tr("All accounts"), QVariant());
	if (d->psi)
	{
		foreach (PsiAccount* account, d->psi->contactList()->enabledAccounts())
			ui_.accountsBox->addItem(IconsetFactory::icon("psi/account").icon(), account->nameWithJid(), QVariant(account->id()));
	}
	//select active account
	ui_.accountsBox->setCurrentIndex(ui_.accountsBox->findData(getCurrentAccountId()));
	//connect signal after the list is populated to prevent execution in the middle of the loop
	connect(ui_.accountsBox, SIGNAL(currentIndexChanged(const QString)), SLOT(changeAccount(const QString)));
}

void HistoryDlg::loadContacts()
{
	QStringList jids;
	ui_.jidList->clear();
	ui_.jidList2->clear();
	ui_.privList->clear();
	ui_.advList->clear();
	ui_.msgLog->clear();
	QList<PsiContact*> contactList;
	QString pa_id = ui_.accountsBox->itemData(ui_.accountsBox->currentIndex()).toString();
	if (pa_id.isEmpty())
		contactList = d->psi->contactList()->contacts();
	else
		contactList = d->psi->contactList()->getAccount(pa_id)->contactList();
	// Roster contacts
	foreach (PsiContact* contact, contactList) {
		if(contact->isConference()
		|| contact->isPrivate())
			continue;
		QString contactId = contact->account()->id() + "|" + contact->jid().bare();
		if (jids.contains(contactId))
			continue;

		QListWidgetItem *item = new QListWidgetItem(contact->name(), ui_.jidList);
		item->setToolTip(makeContactToolTip(contact, true));
		item->setIcon(PsiIconset::instance()->statusPtr(contact->jid(),Status(Status::Online))->icon());
		//item->setIcon(PsiIconset::instance()->status(contact->status()).icon());
		item->setStatusTip(contactId);
		ui_.jidList->addItem(item);
		jids.append(contactId);
		if (contact->jid().bare() == d->jid.bare()) {
			ui_.contactToolBox->setCurrentIndex(0);
			ui_.jidList->setCurrentItem(item);
		}
	}
	// Self contact
	QString currentAccountId = ui_.accountsBox->itemData(ui_.accountsBox->currentIndex()).toString();
	foreach (PsiAccount *pa, d->psi->contactList()->accounts()) {
		if (currentAccountId.isEmpty() || pa->id() == currentAccountId)
		{
			PsiContact* self = pa->selfContact();
			QString contactId = pa->id() + "|" + self->jid().bare();
			if(!jids.contains(contactId)) {
				QListWidgetItem *item = new QListWidgetItem(self->name(), ui_.jidList);
				item->setToolTip(makeContactToolTip(self, true));
				item->setIcon(PsiIconset::instance()->statusPtr(self->jid(),Status(Status::Online))->icon());
				item->setStatusTip(contactId);
				ui_.jidList->addItem(item);
				jids.append(contactId);
				if (self->jid().bare() == d->jid.bare()) {
					ui_.contactToolBox->setCurrentIndex(0);
					ui_.jidList->setCurrentItem(item);
				}
			}
			if (!currentAccountId.isEmpty())
				break;
		}
	}
	// Not in roster list
	foreach (const EDB::ContactItem &ci, d->psi->edb()->contacts(pa_id, EDB::Contact)) {
		QString contactId = ci.accId + "|" + ci.jid.bare();
		if (!jids.contains(contactId)) {
			QListWidgetItem *item = new QListWidgetItem(ci.jid.bare(), ui_.jidList2);
			item->setToolTip(makeContactToolTip(ci.accId, ci.jid, true));
			item->setIcon(PsiIconset::instance()->statusPtr(ci.jid, Status(Status::Offline))->icon());
			item->setStatusTip(contactId);
			ui_.jidList2->addItem(item);
			if (ci.jid.bare() == d->jid.bare()) {
				ui_.contactToolBox->setCurrentIndex(1);
				ui_.jidList2->setCurrentItem(item);
			}
		}
	}
	// Private
	foreach (const EDB::ContactItem &ci, d->psi->edb()->contacts(pa_id, EDB::GroupChatContact)) {
		QString contactId = ci.accId + "|" + ci.jid.full();
		QListWidgetItem *item = new QListWidgetItem(ci.jid.resource(), ui_.privList);
		item->setToolTip(makeContactToolTip(ci.accId, ci.jid, false));
		item->setIcon(PsiIconset::instance()->statusPtr(ci.jid, Status(Status::Offline))->icon());
		item->setStatusTip(contactId);
		ui_.privList->addItem(item);
		if (ci.jid == d->jid) {
			ui_.contactToolBox->setCurrentIndex(2);
			ui_.privList->setCurrentItem(item);
		}
	}
	// Advanced
	{
		QListWidgetItem *item = new QListWidgetItem(tr("All contacts"), ui_.advList);
		item->setToolTip(tr("All contacts"));
		item->setStatusTip(QString());
		ui_.advList->addItem(item);
	}

	ui_.jidList->sortItems();
	ui_.jidList2->sortItems();
	ui_.privList->sortItems();
}

void HistoryDlg::openSelectedContact()
{
	ui_.msgLog->clear();

	QListWidget *contactList = NULL;
	switch (ui_.contactToolBox->currentIndex()) {
		case 0:
			contactList = ui_.jidList;
			break;
		case 1:
			contactList = ui_.jidList2;
			break;
		case 2:
			contactList = ui_.privList;
			break;
	}
	d->pa = 0;
	d->jid = XMPP::Jid();
	if (contactList) {
		QListWidgetItem *item = contactList->currentItem();
		if (item) {
			QString sId = item->statusTip();
			if (!sId.isEmpty())
			{
				d->pa = d->psi->contactList()->getAccount(sId.section('|', 0, 0));
				d->jid = XMPP::Jid(sId.section('|', 1, -1));
			}
		}
	}

	QString sTitle;
	if (!d->jid.isEmpty()) {
		UserListItem *u = currentUserListItem();
		if (u) {
			sTitle = u->name() + " (" + u->jid().full() + ")";
		}
		else {
			sTitle = d->jid.full();
		}
	}
	else {
		QString paId = ui_.accountsBox->itemData(ui_.accountsBox->currentIndex()).toString();
		if (!paId.isEmpty())
			d->pa = d->psi->contactList()->getAccount(paId);
		sTitle = tr("All contacts");
	}
	setWindowTitle(sTitle);
	getLatest();
}

void HistoryDlg::highlightBlocks(const QString text)
{
	QTextCursor cur = ui_.msgLog->textCursor();
	cur.clearSelection();
	cur.movePosition(QTextCursor::Start);
	ui_.msgLog->setTextCursor(cur);

	if (text.isEmpty()) {
		getLatest();
		return;
	}

	QList<QTextEdit::ExtraSelection> extras;
	QTextEdit::ExtraSelection highlight;
	highlight.format.setBackground(Qt::yellow);
	highlight.cursor = ui_.msgLog->textCursor();

	bool found = ui_.msgLog->find(text);
	while (found)
	{
		highlight.cursor = ui_.msgLog->textCursor();
		extras << highlight;
		found = ui_.msgLog->find(text);
	}

	ui_.msgLog->setExtraSelections(extras);
}

void HistoryDlg::findMessages()
{
	//get the oldest event as a starting point
	startRequest();
	d->reqType = TypeFindOldest;
	getEDBHandle()->get(getCurrentAccountId(), d->jid, QDateTime(), EDB::Forward, 0, 1);
}

void HistoryDlg::removeHistory()
{
	int res = QMessageBox::question(this, tr("Remove history"),
					tr("Are you sure you want to completely remove history for a contact %1?").arg(d->jid.full())
					,QMessageBox::Ok | QMessageBox::Cancel);
	if(res == QMessageBox::Ok) {
		getEDBHandle()->erase(getCurrentAccountId(), d->jid);
		openSelectedContact();
	}
}

void HistoryDlg::openChat()
{
	if (d->pa && !d->jid.isEmpty())
		d->pa->actionOpenChat2(d->jid);
}

void HistoryDlg::exportHistory()
{
	UserListItem *u = currentUserListItem();
	QString them;
	if(u) {
		them = JIDUtil::nickOrJid(u->name(), u->jid().full());
	} else {
		//if (d->jid.isEmpty())
		//	return;
		them = d->jid.full();
	}
	QString s = (!them.isEmpty()) ? JIDUtil::encode(them).toLower() : "all_contacts";
	QString fname = FileUtil::getSaveFileName(this,
						  tr("Export message history"),
						  s + ".txt",
						  tr("Text files (*.txt);;All files (*.*)"));
	if(fname.isEmpty())
		return;

	QFile f(fname);
	if(!f.open(QIODevice::WriteOnly | QIODevice::Truncate | QIODevice::Text)) {
		QMessageBox::information(this, tr("Error"), tr("Error writing to file."));
		return;
	}
	QTextStream stream(&f);

	int start = 0;
	startRequest();
	QString paId = getCurrentAccountId();
	int max = 0;
	{
		quint64 edbCnt = d->psi->edb()->eventsCount(paId, d->jid);
		if (edbCnt > 1000) {
			max = edbCnt / 1000;
			if ((edbCnt % 1000) != 0)
				++max;
			showProgress(max);
		}
	}
	while(1) {
		EDBHandle *h = new EDBHandle(d->psi->edb());
		h->get(paId, d->jid, QDateTime(), EDB::Forward, start, 1000);
		while(h->busy()) {
			qApp->processEvents();
		}

		const EDBResult r = h->result();
		int cnt = r.count();

		// events are in forward order
		for(int i = 0; i < cnt; ++i) {
			EDBItemPtr item = r.value(i);
			PsiEvent::Ptr e(item->event());
			QString txt;

			QString ts = e->timeStamp().toString(Qt::LocalDate);

			QString nick;
			if(e->originLocal()) {
				if (e->account())
					nick = e->account()->nick();
				else
					nick = tr("deleted");
			}
			else {
				if (!them.isEmpty())
					nick = them;
				else
					nick = e->from().full();
			}

			if(e->type() == PsiEvent::Message) {
				MessageEvent::Ptr me = e.staticCast<MessageEvent>();
				stream << QString("[%1] <%2>: ").arg(ts, nick)/* << endl*/;

				QStringList lines = me->message().body().split('\n', QString::KeepEmptyParts);
				foreach(const QString& str, lines) {
					QStringList sub = wrapString(str, 72);
					foreach(const QString& str2, sub) {
						txt += str2 + "\n" + QString("    ");
					}
				}
			}
			else {
				continue;
			}

			stream << txt << endl;
		}
		delete h;

		if (max > 0)
			incrementProgress();

		// done!
		if(cnt == 0)
			break;

		start += 1000;
	}
	f.close();
	stopRequest();
}

void HistoryDlg::doMenu()
{
	openSelectedContact();
	bool fAll = (!d->pa || d->jid.isEmpty());
	QMenu *m = new QMenu();
	if (!fAll)
		m->addAction(IconsetFactory::icon("psi/chat").icon(), tr("&Open chat"), this, SLOT(openChat()));
	m->addAction(IconsetFactory::icon("psi/save").icon(), tr("&Export history"), this, SLOT(exportHistory()));
	if (!fAll)
		m->addAction(IconsetFactory::icon("psi/clearChat").icon(), tr("&Delete history"), this, SLOT(removeHistory()));
	m->exec(QCursor::pos());
	delete m;
}

void HistoryDlg::edbFinished()
{
	setButtons(false);
	stopRequest();

	EDBHandle* h = qobject_cast<EDBHandle*>(sender());
	if(!h) {
		return;
	}

	const EDBResult r = h->result();
	if (h->lastRequestType() == EDBHandle::Read)
	{
		if (r.count() > 0)
		{
			if (d->reqType == TypeLatest || d->reqType == TypePrevious)
			{
				// events are in backward order
				d->begin_row = h->beginRow();
				d->end_row = d->begin_row + r.count() - 1;

				d->can_backward = (d->begin_row > 0);
				d->can_forward = (d->reqType != TypeLatest);
				displayResult(r, (d->reqType == TypeLatest) ? EDB::Forward : EDB::Backward);
				setButtons();
			}
			else if (d->reqType == TypeEarliest || d->reqType == TypeNext || d->reqType == TypeDate)
			{
				// events are in forward order
				d->begin_row = h->beginRow();
				d->end_row = d->begin_row + r.count() - 1;

				d->can_backward = (d->begin_row > 0);
				d->can_forward = true;
				displayResult(r, EDB::Backward);
				setButtons();
			}
			else if (d->reqType == TypeFindOldest)
			{
				QString str = ui_.searchField->text();
				if (str.isEmpty())
				{
					getLatest();
				}
				else
				{
					d->reqType = TypeFind;
					d->findStr = str;
					startRequest();
					getEDBHandle()->find(getCurrentAccountId(), str, d->jid, QDateTime(), EDB::Forward);
					setButtons();
				}
			}
			else if (d->reqType == TypeFind)
			{
				d->begin_row = 0;
				d->end_row = r.count() - 1;
				displayResult(r, EDB::Forward);
				highlightBlocks(ui_.searchField->text());
			}

		}
		else
		{
			// no more data. TODO: visualization is needed.
			bool buttons_updated = false;
			if (d->reqType == TypePrevious)
			{
				d->can_backward = false;
				buttons_updated = true;
			}
			else if (d->reqType == TypeLatest || d->reqType == TypeEarliest)
			{
				d->can_backward = false;
				d->can_forward = false;
				buttons_updated = true;
			}
			else if (d->reqType == TypeNext || d->reqType == TypeDate)
			{
				d->can_forward = false;
				buttons_updated = true;
			}
			if (d->reqType == TypeFind)
				ui_.msgLog->clear();

			if (buttons_updated)
				setButtons();
		}
	}
	delete h;
}

void HistoryDlg::setButtons()
{
	ui_.buttonPrevious->setEnabled(d->can_backward);
	ui_.buttonNext->setEnabled(d->can_forward);
	ui_.buttonEarliest->setEnabled(d->can_backward);
	ui_.buttonLastest->setEnabled(d->can_forward);
}

void HistoryDlg::setButtons(bool act)
{
	ui_.buttonPrevious->setEnabled(act);
	ui_.buttonNext->setEnabled(act);
	ui_.buttonEarliest->setEnabled(act);
	ui_.buttonLastest->setEnabled(act);
}

void HistoryDlg::refresh()
{
	ui_.calendar->setSelectedDate(QDate::currentDate());
	ui_.searchField->clear();
	getLatest();
}

void HistoryDlg::getLatest()
{
	d->reqType = TypeLatest;
	startRequest();
	getEDBHandle()->get(getCurrentAccountId(), d->jid, QDateTime(), EDB::Backward, 0, 50);
}

void HistoryDlg::getEarliest()
{
	d->reqType = TypeEarliest;
	startRequest();
	getEDBHandle()->get(getCurrentAccountId(), d->jid, QDateTime(), EDB::Forward, 0, 50);
}

void HistoryDlg::getPrevious()
{
	d->reqType = TypePrevious;
	ui_.buttonPrevious->setEnabled(false);
	int begin = (d->begin_row < 50) ? 0 : d->begin_row - 50;
	getEDBHandle()->get(getCurrentAccountId(), d->jid, QDateTime(), EDB::Forward, begin, 50);
}

void HistoryDlg::getNext()
{
	d->reqType = TypeNext;
	ui_.buttonNext->setEnabled(false);
	getEDBHandle()->get(getCurrentAccountId(), d->jid, QDateTime(), EDB::Forward, d->end_row + 1, 50);
}

void HistoryDlg::getDate()
{
	const QDate date = ui_.calendar->selectedDate();
	d->reqType = TypeDate;
	d->date = date;
	QDateTime first(d->date);
	startRequest();
	getEDBHandle()->get(getCurrentAccountId(), d->jid, first, EDB::Forward, 0, 50);
}

void HistoryDlg::removedContact(PsiContact *pc)
{
	QString contactId = pc->account()->id() + "|" + pc->jid().bare().toLower();
	QString curId;
	QListWidgetItem *lwi = ui_.jidList->currentItem();
	if (lwi)
		curId = lwi->statusTip();
	for(int i = 0; i < ui_.jidList->count(); i++) {
		QListWidgetItem *it = ui_.jidList->item(i);
		if(it && it->statusTip() == contactId) {
			ui_.jidList->removeItemWidget(it);
			it = new QListWidgetItem(pc->jid().bare(), ui_.jidList2);
			it->setToolTip(makeContactToolTip(pc, true));
			it->setIcon(PsiIconset::instance()->statusPtr(pc->jid().bare(), Status(Status::Offline))->icon());
			it->setStatusTip(contactId);
			ui_.jidList2->addItem(it);
			if(contactId == curId) {
				ui_.jidList->setCurrentRow(0);
				if (ui_.contactToolBox->currentIndex() == 0) {
					ui_.jidList2->setCurrentItem(it);
					ui_.contactToolBox->setCurrentIndex(1);
				}
			}
			break;
		}
	}
}

void HistoryDlg::optionUpdated(const QString &option)
{
#ifndef HAVE_X11
	if(option == "options.ui.automatically-copy-selected-text") {
		d->autoCopyText = PsiOptions::instance()->getOption(option).toBool();
	} else
#endif
	if(option == "options.ui.look.colors.messages.sent")
			d->sentColor = PsiOptions::instance()->getOption(option).toString();
	else if(option == "options.ui.look.colors.messages.received")
			d->receivedColor = PsiOptions::instance()->getOption(option).toString();
	else if(option == "options.ui.emoticons.use-emoticons")
			d->emoticons  = PsiOptions::instance()->getOption(option).toBool();
	else if(option == "options.ui.chat.legacy-formatting")
			d->formatting = PsiOptions::instance()->getOption(option).toBool();
}
#ifndef HAVE_X11
void HistoryDlg::autoCopy()
{
	if (d->autoCopyText) {
		ui_.msgLog->copy();
	}
}
#endif
void HistoryDlg::displayResult(const EDBResult &r, int direction, int max)
{
	int i  = (direction == EDB::Forward) ? r.count() - 1 : 0;
	int at = 0;
	ui_.msgLog->clear();
	bool fAll = d->jid.isEmpty();
	while (i >= 0 && i <= r.count() - 1 && (max == -1 ? true : at < max))
	{
		EDBItemPtr item = r.value(i);
		PsiEvent::Ptr e(item->event());
		if (e->type() == PsiEvent::Message) {
			PsiAccount *pa = e->account();
			QString from;
			if (pa)
			{
				UserListItem *u = pa->findFirstRelevant(e->from().full());
				if(u) {
					if (!u->name().trimmed().isEmpty())
						from = u->name().trimmed();
				}
			}
			if (from.isEmpty())
			{
				from = e->from().resource().trimmed(); // for offline conferences
				if (from.isEmpty())
						from = e->from().node();
			}
			MessageEvent::Ptr me = e.staticCast<MessageEvent>();
			QString msg = me->message().body();
			msg = TextUtil::linkify(TextUtil::plain2rich(msg));

			if (d->emoticons)
				msg = TextUtil::emoticonify(msg);
			if (d->formatting)
				msg = TextUtil::legacyFormat(msg);

			if (me->originLocal())
			{
				QString nick = (pa) ? TextUtil::plain2rich(pa->nick()) : tr("deleted");
				msg = "<span style='color:" + d->sentColor + "'>" + me->timeStamp().toString("[dd.MM.yyyy hh:mm:ss]") + " &lt;" + nick
					+ ((fAll) ? QString(" -> %1").arg(TextUtil::plain2rich(from)) : QString())
					+ "&gt; " + msg + "</span>";
			}
			else
				msg = "<span style='color:" + d->receivedColor + "'>" + me->timeStamp().toString("[dd.MM.yyyy hh:mm:ss]") + " &lt;" +  TextUtil::plain2rich(from) + "&gt; " + msg + "</span>";

			ui_.msgLog->appendText(msg);
		}

		++at;
		i += (direction == EDB::Forward) ? -1 : +1;
	}

	ui_.msgLog->verticalScrollBar()->setValue(ui_.msgLog->verticalScrollBar()->maximum());
}

UserListItem* HistoryDlg::currentUserListItem() const
{
	UserListItem* u = 0;
	if (d->pa && !d->jid.isEmpty()) {
		u = d->pa->findFirstRelevant(d->jid);
	}
	return u;
}

void HistoryDlg::startRequest()
{
	if(!ui_.busy->isActive()) {
		ui_.busy->start();
	}
	setEnabled(false);
}

void HistoryDlg::stopRequest()
{
	if(ui_.busy->isActive()) {
		ui_.busy->stop();
	}
	ui_.progressBar->setVisible(false);
	setEnabled(true);
#ifdef Q_OS_MAC
	// To workaround a Qt bug
	// https://bugreports.qt-project.org/browse/QTBUG-26351
	setFocus();
#endif
}

void HistoryDlg::showProgress(int max)
{
	ui_.progressBar->setValue(0);
	ui_.progressBar->setMaximum(max);
	ui_.progressBar->setVisible(true);
}

void HistoryDlg::incrementProgress()
{
	ui_.progressBar->setValue(ui_.progressBar->value() + 1);
}

EDBHandle* HistoryDlg::getEDBHandle()
{
	EDBHandle* h = new EDBHandle(d->psi->edb());
	connect(h, SIGNAL(finished()), SLOT(edbFinished()));
	return h;
}

QString HistoryDlg::makeContactToolTip(const PsiContact *contact, bool bare) const
{
	QString jidStr = JIDUtil::toString(contact->jid(), !bare);
	if (ui_.accountsBox->itemData(ui_.accountsBox->currentIndex()).isNull())
		jidStr.append(QString(" [%1]").arg(contact->account()->name()));
	return jidStr;
}

QString HistoryDlg::makeContactToolTip(const QString &accId, const XMPP::Jid &jid, bool bare) const
{
	QString jidStr = JIDUtil::toString(jid, !bare);
	if (ui_.accountsBox->itemData(ui_.accountsBox->currentIndex()).isNull()) {
		PsiAccount *pa = d->psi->contactList()->getAccount(accId);
		jidStr.append(QString(" [%1]").arg((pa) ? pa->name() : tr("deleted")));
	}
	return jidStr;
}

QString HistoryDlg::getCurrentAccountId() const
{
	if (d->pa)
		return d->pa->id();
	return QString();
}
